# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_queries.ipynb.

# %% auto 0
__all__ = ['AliasedQuery', 'Table', 'QueryBase', 'Joiner', 'Selector', 'SelectQuery', 'UnionQuery', 'Query', 'Exists']

# %% ../nbs/04_queries.ipynb 2
from .utils import execute
from .terms import Field
import pikaQ.functions as fn

# %% ../nbs/04_queries.ipynb 4
class Table:
    """A class with star and as_ methods to be used as a Table/AliasesQuery in a query. Any other attribute will be treated as a field."""
    def __init__(self, name) -> None:
        self.name = self.alias = name
        self.get_sql = self.execute

    def as_(self, alias):
        self.alias = alias
        return self
    
    def star(self):
        return Field(f"{self.alias}.*")

    def __getattr__(self, __name: str):
        if __name.startswith('__'):
            raise AttributeError
        else:
            # if an alias is assigned, always use it as the table name when accessing a field
            return Field(f"{self.alias}.{__name}")

    def quoted_name(self, quote_char):
        name_list = self.name.split('.')
        name = '.'.join([f"{quote_char}{n}{quote_char}" for n in name_list])
        return name
    
    def execute(self, **kwargs):
        q = kwargs.get('quote_char', '') or ''
        quoted_name = self.quoted_name(q)
        if self.alias != self.name:
            return f"{quoted_name} as {q}{self.alias}{q}"
        else:
            return quoted_name


AliasedQuery = Table

# %% ../nbs/04_queries.ipynb 7
class QueryBase:
    """An empty query class to be inherited by all query classes. A convenient tool to make all query objects belong to this same class."""
    pass


class Joiner(QueryBase):
    """Join clause has to be followed by an on clause"""
    def __init__(self, select_query, query, how=None) -> None:
        self.select_query = select_query
        self.query = query
        self.how = how

    def on(self, condition):
        l = self.select_query.dic.get('join', [])
        l.append((self.query, self.how, condition))
        self.select_query.dic['join'] = l
        return self.select_query


class Selector(QueryBase):
    """Select clause could be followed by a distinct clause"""
    def __init__(self, select_query, *args) -> None:
        self.select_query = select_query
        self.select_query.dic['select'] = {"columns": list(args)}
        # add parse_select method to the select_query object
        self.select_query.parse_select = self.parse_select

    def __getattr__(self, __name: str):
        # selector inherits all methods from the SelectQuery object
        # so that the distinct method is optional
        # executing some of these methods can get back to the select_query object
        return getattr(self.select_query, __name)

    def parse_select(self, **kwargs):
        key = 'select'
        dic = self.dic[key]
        columns = dic['columns']
        columns = [execute(column, **kwargs) for column in columns]
        if dic.get('distinct'):
            return f"{key} distinct " + ', '.join(columns)
        else:
            return f"{key} " + ', '.join(columns)

    def distinct(self):
        self.select_query.dic['select']['distinct'] = True
        return self.select_query


class SelectQuery(QueryBase):
    """The class to construct a select query. It returns a `Joiner` object when called with the method join. It returns a `Selector` object when called with the method select."""
    keys_simple = ['from', 'groupby', 'orderby', 'where', 'having', 'limit']
    # the order to put together the final sql query
    sql_keys = ['with', 'select', 'from', 'join', 'where', 'groupby', 'having', 'orderby', 'limit']
    key_translation = {
        'groupby': 'group by',
        'orderby': 'order by'
    }

    def __init__(self) -> None:
        # the dictionary to store the queries for each clause
        self.dic = {}
        self.get_sql = self.execute

    def join(self, query, how=None):
        return Joiner(self, query, how)

    def select(self, *args):
        return Selector(self, *args)

    def __getattr__(self, __name):
        # Construct the corresponding method functions
        # remove the leading and trailing underscores
        __name = __name.strip('_').lower()
        # get the string after the first underscore
        key = '_'.join(__name.split('_')[1:])

        if __name.startswith('parse'):
            # Construct parse_xxx functions
            def inner(**kwargs):
                if key in ['from', 'where', 'having', 'limit']:
                    q = self.dic[key]
                    if isinstance(q, QueryBase):
                        return f"{self.key_translation.get(key, key)} ({execute(q, **kwargs)})"
                    else:
                        return f"{self.key_translation.get(key, key)} {execute(q, **kwargs)}"
                elif key in ['groupby', 'orderby']:
                    args = self.dic[key]
                    args = [execute(arg, **kwargs) for arg in args]
                    return f"{self.key_translation.get(key, key)} " + ', '.join(args)
                elif key == 'with':
                    s = self.dic[key]
                    qq = [f"{a} as (\n{execute(q, **kwargs)})" for q, a in s]
                    return f"{key} " + "\n\n, ".join(qq) + "\n"
                elif key == 'join':
                    l = self.dic[key]
                    parsed = []
                    for q, how, cond in l:
                        if how:
                            how = how + ' '
                        else:
                            how = ''
                        if isinstance(q, QueryBase):
                            sub_q = f"({execute(q, **kwargs)})"
                        else:
                            sub_q = execute(q, **kwargs)
                        parsed.append(f"{how}join {sub_q} on {execute(cond, **kwargs)}")
                    return '\n'.join(parsed)
                else:
                    raise AttributeError(f"parsing function for {key} is not defined!")
        else:
            # Construct other method functions
            if __name in ['from', 'where', 'having', 'limit']:
                def inner(q):
                    self.dic[__name] = q
                    return self
            elif __name in ['groupby', 'orderby']:
                def inner(*args):
                    self.dic[__name] = args
                    return self
            elif __name == 'with':
                def inner(query, alias):
                    l = self.dic.get('with', [])
                    l.append((query, alias))
                    self.dic['with'] = l
                    return self
            else:
                raise AttributeError(f"{__name} is not a valid attribute!")

        return inner

    def execute(self, **kwargs):
        dic_sql = {}

        for k, v in self.dic.items():
            dic_sql[k] = getattr(self, f"parse_{k}")(**kwargs)

        if missing_keys := [key for key in dic_sql if key not in self.sql_keys]:
            raise ValueError(f"The order of {missing_keys} in the final SQL are not specified in self.sql_keys!")

        sql = '\n'.join([dic_sql[key] for key in self.sql_keys if key in dic_sql])
        return sql
    
    def union(self, query):
        if not isinstance(query, QueryBase):
            raise TypeError(f"{query} is not an instance of QueryBase!")
        return UnionQuery(self, query, union_type='UNION')

    def __add__(self, query):
        return self.union(query)

    def union_all(self, query):
        if not isinstance(query, QueryBase):
            raise TypeError(f"{query} is not an instance of QueryBase!")
        return UnionQuery(self, query, union_type='UNION ALL')

    def __mul__(self, query):
        return self.union_all(query)


class UnionQuery(QueryBase):
    """The class to construct a union query."""
    def __init__(self, q1, q2, union_type='UNION') -> None:
        self.q1 = q1
        self.q2 = q2
        self.union_type = union_type

    def execute(self, **kwargs):
        return f"{execute(self.q1, **kwargs)} {self.union_type} {execute(self.q2, **kwargs)}"


class Query(QueryBase):
    """The main class to construct a query. It returns a `SelectQuery` object when called with the classmethod from_ or with_. One can extend this class to add more classmethods to construct different types of queries."""
    q = SelectQuery

    @classmethod
    def from_(cls, query):
        q = cls.q()
        q.dic['from'] = query
        return q

    @classmethod
    def with_(cls, query, alias):
        q = cls.q()
        q.dic['with'] = [(query, alias)]
        return q

# %% ../nbs/04_queries.ipynb 13
class Exists:
    """Exists statement"""
    def __init__(self, query:Query) -> None:
        super().__init__()
        self.query = query
        self.get_sql = self.execute

    def execute(self, **kwargs):
        return f"EXISTS ({self.query.get_sql(**kwargs)})"



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_functions.ipynb.

# %% auto 0
__all__ = ['Max', 'Min', 'Sum', 'Avg', 'Count', 'Abs', 'Round', 'First', 'Last', 'MonthsBetween', 'RowNumber', 'Rank',
           'DenseRank', 'PercentRank', 'CumeDist', 'Ntile', 'Lag', 'Lead', 'FirstValue', 'LastValue', 'NthValue',
           'CustomFunction', 'Cast', 'Coalesce', 'Concat', 'convert_date_format', 'Date', 'AddMonths', 'DateDiff',
           'DateTrunc']

# %% ../nbs/03_functions.ipynb 2
from fastcore.all import patch, delegates
from .utils import execute
from .terms import FieldBase, Field, custom_func, OverClause

# %% ../nbs/03_functions.ipynb 4
class CustomFunction(FieldBase):
    """A convenient class for creating custom functions."""

    def __init__(self, 
                 func_name: str,    # name of the function 
                 arg_names: list,   # list of arg names
                 window_func=False,  # whether this can be used as a window function
                 distinct_option=False # whether this function can be used with the distinct option
        ) -> None:
        import inspect
        super().__init__()
        self.func_name = func_name
        self.arg_names = arg_names
        self.window_func = window_func
        self.distinct_option = distinct_option
        self.distinct_ = False  # whether the distinct option is used
        self.get_sql = self.execute
        self.__qualname__ = func_name
        self.__signature__ = inspect.Signature(parameters=[inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in arg_names]) 
        self.__doc__ = None

    def __call__(self, *args):
        if len(args) != len(self.arg_names):
            raise ValueError(f"The number of args provided {len(args)} is not the same as the number of args expected by this function ({len(self.arg_names)})!")
        def func(*args):
            if self.distinct_:
                return f"{self.func_name}(DISTINCT {', '.join(args)})"
            else:
                return f"{self.func_name}({', '.join(args)})"

        self.func = func
        self.args = args
        return self

    def execute(self, **kwargs):
        args = [str(execute(arg, **kwargs)) for arg in self.args]
        return self.func(*args)

    def distinct(self):
        if self.distinct_option==True:
            self.distinct_ = True
            return self
        else:
            raise ValueError(f"This function does not support the distinct option!!")


@patch
def over(self:CustomFunction, partition_by):
    if self.window_func==True:
        return OverClause(self).over(partition_by)
    else:
        raise ValueError(f"This function is not a window function!!")

# %% ../nbs/03_functions.ipynb 11
Max = CustomFunction('MAX', ['field'], window_func=True, distinct_option=True)
Min = CustomFunction('MIN', ['field'], window_func=True, distinct_option=True)
Sum = CustomFunction('SUM', ['field'], window_func=True, distinct_option=True)
Avg = CustomFunction('AVG', ['field'], window_func=True, distinct_option=True)
Count = CustomFunction('COUNT', ['field'], window_func=True, distinct_option=True)
Abs = CustomFunction('ABS', ['field'])
Round = CustomFunction('ROUND', ['field', 'decimals'])
First = CustomFunction('FIRST', ['field'])
Last = CustomFunction('LAST', ['field'])

# %% ../nbs/03_functions.ipynb 21
@custom_func
def Cast(field, type):
    return f"CAST({field} AS {type})"

@custom_func
def Coalesce(*args):
    return f"COALESCE({', '.join(args)})"

@custom_func
def Concat(*args):
    return f"CONCAT({', '.join(args)})"


# date functions
def convert_date_format(format, dialect='sql'):
    if dialect == 'spark':
        return (format.replace('YYYY', 'yyyy')
                .replace('YY', 'yy')
                .replace('DD', 'dd')
                )
    elif dialect == 'snowflake':
        return (format.replace('HH', 'HH24')
                .replace('mm', 'MI')
        )
    else:
        return format

@custom_func
def Date(expression, format=None, dialect='sql'):
    if format:
        format = convert_date_format(format, dialect)
        return f"DATE('{expression}', '{format}')"
    else:
        return f"DATE('{expression}')"

@custom_func
def AddMonths(date, months, dialect='sql'):
    if dialect in ['snowflake', 'spark']:
        return f"ADD_MONTHS({date}, {months})"
    elif dialect == 'athena':
        return f"date_add('month', {months}, {date})"
    else:
        return f"DATE_ADD({date}, INTERVAL {months} MONTH)"

@custom_func
def DateDiff(interval, start_date, end_date, dialect='sql'):
    if dialect == 'athena':
        return f"date_diff('{interval}', {start_date}, {end_date})"
    else:
        return f"DATEDIFF({interval}, {start_date}, {end_date})"

@custom_func
def DateTrunc(interval, date, dialect='sql'):
    if dialect == 'spark':
        return f"TRUNC({date}, '{interval}')"
    else:
        return f"DATE_TRUNC('{interval}', {date})"


# %% ../nbs/03_functions.ipynb 22
MonthsBetween = CustomFunction('MONTHS_BETWEEN', ['start_date', 'end_date'])

# Window Functions
RowNumber = CustomFunction('ROW_NUMBER', [], window_func=True)
Rank = CustomFunction('RANK', [], window_func=True)
DenseRank = CustomFunction('DENSE_RANK', [], window_func=True)
PercentRank = CustomFunction('PERCENT_RANK', [], window_func=True)
CumeDist = CustomFunction('CUME_DIST', [], window_func=True)
Ntile = CustomFunction('NTILE', ['num_buckets'], window_func=True)
Lag = CustomFunction('LAG', ['field', 'offset'], window_func=True)
Lead = CustomFunction('LEAD', ['field', 'offset'], window_func=True)
FirstValue = CustomFunction('FIRST_VALUE', ['field'], window_func=True)
LastValue = CustomFunction('LAST_VALUE', ['field'], window_func=True)
NthValue = CustomFunction('NTH_VALUE', ['field', 'n'], window_func=True)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/pikaQ',
                'doc_host': 'https://feynlee.github.io',
                'git_url': 'https://github.com/feynlee/pikaQ',
                'lib_path': 'pikaQ'},
  'syms': { 'pikaQ.functions': { 'pikaQ.functions.AddMonths': ('functions.html#addmonths', 'pikaQ/functions.py'),
                                 'pikaQ.functions.Cast': ('functions.html#cast', 'pikaQ/functions.py'),
                                 'pikaQ.functions.Coalesce': ('functions.html#coalesce', 'pikaQ/functions.py'),
                                 'pikaQ.functions.Concat': ('functions.html#concat', 'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction': ('functions.html#customfunction', 'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction.__call__': ( 'functions.html#customfunction.__call__',
                                                                              'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction.__init__': ( 'functions.html#customfunction.__init__',
                                                                              'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction.distinct': ( 'functions.html#customfunction.distinct',
                                                                              'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction.execute': ('functions.html#customfunction.execute', 'pikaQ/functions.py'),
                                 'pikaQ.functions.CustomFunction.over': ('functions.html#customfunction.over', 'pikaQ/functions.py'),
                                 'pikaQ.functions.Date': ('functions.html#date', 'pikaQ/functions.py'),
                                 'pikaQ.functions.DateDiff': ('functions.html#datediff', 'pikaQ/functions.py'),
                                 'pikaQ.functions.DateTrunc': ('functions.html#datetrunc', 'pikaQ/functions.py'),
                                 'pikaQ.functions.convert_date_format': ('functions.html#convert_date_format', 'pikaQ/functions.py')},
            'pikaQ.queries': { 'pikaQ.queries.Exists': ('queries.html#exists', 'pikaQ/queries.py'),
                               'pikaQ.queries.Exists.__init__': ('queries.html#exists.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Exists.execute': ('queries.html#exists.execute', 'pikaQ/queries.py'),
                               'pikaQ.queries.Joiner': ('queries.html#joiner', 'pikaQ/queries.py'),
                               'pikaQ.queries.Joiner.__init__': ('queries.html#joiner.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Joiner.on': ('queries.html#joiner.on', 'pikaQ/queries.py'),
                               'pikaQ.queries.Query': ('queries.html#query', 'pikaQ/queries.py'),
                               'pikaQ.queries.Query.from_': ('queries.html#query.from_', 'pikaQ/queries.py'),
                               'pikaQ.queries.Query.with_': ('queries.html#query.with_', 'pikaQ/queries.py'),
                               'pikaQ.queries.QueryBase': ('queries.html#querybase', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery': ('queries.html#selectquery', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.__add__': ('queries.html#selectquery.__add__', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.__getattr__': ('queries.html#selectquery.__getattr__', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.__init__': ('queries.html#selectquery.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.__mul__': ('queries.html#selectquery.__mul__', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.execute': ('queries.html#selectquery.execute', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.join': ('queries.html#selectquery.join', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.select': ('queries.html#selectquery.select', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.union': ('queries.html#selectquery.union', 'pikaQ/queries.py'),
                               'pikaQ.queries.SelectQuery.union_all': ('queries.html#selectquery.union_all', 'pikaQ/queries.py'),
                               'pikaQ.queries.Selector': ('queries.html#selector', 'pikaQ/queries.py'),
                               'pikaQ.queries.Selector.__getattr__': ('queries.html#selector.__getattr__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Selector.__init__': ('queries.html#selector.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Selector.distinct': ('queries.html#selector.distinct', 'pikaQ/queries.py'),
                               'pikaQ.queries.Selector.parse_select': ('queries.html#selector.parse_select', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table': ('queries.html#table', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.__getattr__': ('queries.html#table.__getattr__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.__init__': ('queries.html#table.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.as_': ('queries.html#table.as_', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.execute': ('queries.html#table.execute', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.quoted_name': ('queries.html#table.quoted_name', 'pikaQ/queries.py'),
                               'pikaQ.queries.Table.star': ('queries.html#table.star', 'pikaQ/queries.py'),
                               'pikaQ.queries.UnionQuery': ('queries.html#unionquery', 'pikaQ/queries.py'),
                               'pikaQ.queries.UnionQuery.__init__': ('queries.html#unionquery.__init__', 'pikaQ/queries.py'),
                               'pikaQ.queries.UnionQuery.execute': ('queries.html#unionquery.execute', 'pikaQ/queries.py')},
            'pikaQ.terms': { 'pikaQ.terms.ArithmeticExpression': ('terms.html#arithmeticexpression', 'pikaQ/terms.py'),
                             'pikaQ.terms.ArithmeticExpression.__init__': ('terms.html#arithmeticexpression.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.ArithmeticExpression.execute': ('terms.html#arithmeticexpression.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.ArithmeticExpression.left_needs_parens': ( 'terms.html#arithmeticexpression.left_needs_parens',
                                                                                     'pikaQ/terms.py'),
                             'pikaQ.terms.ArithmeticExpression.over': ('terms.html#arithmeticexpression.over', 'pikaQ/terms.py'),
                             'pikaQ.terms.ArithmeticExpression.right_needs_parens': ( 'terms.html#arithmeticexpression.right_needs_parens',
                                                                                      'pikaQ/terms.py'),
                             'pikaQ.terms.Case': ('terms.html#case', 'pikaQ/terms.py'),
                             'pikaQ.terms.Case.__init__': ('terms.html#case.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Case.check_prev': ('terms.html#case.check_prev', 'pikaQ/terms.py'),
                             'pikaQ.terms.Case.else_': ('terms.html#case.else_', 'pikaQ/terms.py'),
                             'pikaQ.terms.Case.execute': ('terms.html#case.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.Case.when': ('terms.html#case.when', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria': ('terms.html#criteria', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.__and__': ('terms.html#criteria.__and__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.__init__': ('terms.html#criteria.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.__or__': ('terms.html#criteria.__or__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.compose_criteria': ('terms.html#criteria.compose_criteria', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.execute': ('terms.html#criteria.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.negate': ('terms.html#criteria.negate', 'pikaQ/terms.py'),
                             'pikaQ.terms.Criteria.resolve': ('terms.html#criteria.resolve', 'pikaQ/terms.py'),
                             'pikaQ.terms.DelayedFunc': ('terms.html#delayedfunc', 'pikaQ/terms.py'),
                             'pikaQ.terms.DelayedFunc.__init__': ('terms.html#delayedfunc.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.DelayedFunc.execute': ('terms.html#delayedfunc.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.DelayedFunc.over': ('terms.html#delayedfunc.over', 'pikaQ/terms.py'),
                             'pikaQ.terms.Field': ('terms.html#field', 'pikaQ/terms.py'),
                             'pikaQ.terms.Field.__init__': ('terms.html#field.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Field.execute': ('terms.html#field.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.Field.quoted_name': ('terms.html#field.quoted_name', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase': ('terms.html#fieldbase', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__add__': ('terms.html#fieldbase.__add__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__eq__': ('terms.html#fieldbase.__eq__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__ge__': ('terms.html#fieldbase.__ge__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__gt__': ('terms.html#fieldbase.__gt__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__le__': ('terms.html#fieldbase.__le__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__lt__': ('terms.html#fieldbase.__lt__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__mul__': ('terms.html#fieldbase.__mul__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__ne__': ('terms.html#fieldbase.__ne__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__radd__': ('terms.html#fieldbase.__radd__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__rmul__': ('terms.html#fieldbase.__rmul__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__rsub__': ('terms.html#fieldbase.__rsub__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__rtruediv__': ('terms.html#fieldbase.__rtruediv__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__sub__': ('terms.html#fieldbase.__sub__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.__truediv__': ('terms.html#fieldbase.__truediv__', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.ilike': ('terms.html#fieldbase.ilike', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.isin': ('terms.html#fieldbase.isin', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.isnull': ('terms.html#fieldbase.isnull', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.like': ('terms.html#fieldbase.like', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.not_ilike': ('terms.html#fieldbase.not_ilike', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.not_like': ('terms.html#fieldbase.not_like', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.notin': ('terms.html#fieldbase.notin', 'pikaQ/terms.py'),
                             'pikaQ.terms.FieldBase.notnull': ('terms.html#fieldbase.notnull', 'pikaQ/terms.py'),
                             'pikaQ.terms.Following': ('terms.html#following', 'pikaQ/terms.py'),
                             'pikaQ.terms.Following.__init__': ('terms.html#following.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Following.execute': ('terms.html#following.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.NullValue': ('terms.html#nullvalue', 'pikaQ/terms.py'),
                             'pikaQ.terms.NullValue.__init__': ('terms.html#nullvalue.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.NullValue.execute': ('terms.html#nullvalue.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause': ('terms.html#overclause', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.__init__': ('terms.html#overclause.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause._check_rows_or_range': ( 'terms.html#overclause._check_rows_or_range',
                                                                              'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause._resolve_over_statement': ( 'terms.html#overclause._resolve_over_statement',
                                                                                 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.as_': ('terms.html#overclause.as_', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.execute': ('terms.html#overclause.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.orderby': ('terms.html#overclause.orderby', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.over': ('terms.html#overclause.over', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.range': ('terms.html#overclause.range', 'pikaQ/terms.py'),
                             'pikaQ.terms.OverClause.rows': ('terms.html#overclause.rows', 'pikaQ/terms.py'),
                             'pikaQ.terms.Preceding': ('terms.html#preceding', 'pikaQ/terms.py'),
                             'pikaQ.terms.Preceding.__init__': ('terms.html#preceding.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Preceding.execute': ('terms.html#preceding.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.Term': ('terms.html#term', 'pikaQ/terms.py'),
                             'pikaQ.terms.Term.__init__': ('terms.html#term.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Term.as_': ('terms.html#term.as_', 'pikaQ/terms.py'),
                             'pikaQ.terms.Value': ('terms.html#value', 'pikaQ/terms.py'),
                             'pikaQ.terms.Value.__init__': ('terms.html#value.__init__', 'pikaQ/terms.py'),
                             'pikaQ.terms.Value.execute': ('terms.html#value.execute', 'pikaQ/terms.py'),
                             'pikaQ.terms.custom_func': ('terms.html#custom_func', 'pikaQ/terms.py'),
                             'pikaQ.terms.kwargs_func': ('terms.html#kwargs_func', 'pikaQ/terms.py')},
            'pikaQ.utils': { 'pikaQ.utils.delegates': ('utils.html#delegates', 'pikaQ/utils.py'),
                             'pikaQ.utils.execute': ('utils.html#execute', 'pikaQ/utils.py'),
                             'pikaQ.utils.quote_symbol': ('utils.html#quote_symbol', 'pikaQ/utils.py'),
                             'pikaQ.utils.to_sql': ('utils.html#to_sql', 'pikaQ/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_terms.ipynb.

# %% auto 0
__all__ = ['CURRENT_ROW', 'Term', 'Value', 'NullValue', 'FieldBase', 'Field', 'ArithmeticExpression', 'Criteria', 'kwargs_func',
           'DelayedFunc', 'custom_func', 'OverClause', 'Preceding', 'Following', 'Case']

# %% ../nbs/02_terms.ipynb 2
import inspect
from functools import partial
from fastcore.foundation import patch
from .utils import execute, delegates

# %% ../nbs/02_terms.ipynb 4
class Term:
    def __init__(self) -> None:
        self.alias = None
    
    def as_(self, alias):
        self.alias = alias
        return self


class Value(Term):
    """A simple wrapper for a value that will be used in a query."""
    def __init__(self, value) -> None:
        super().__init__()
        self.value = value
        self.get_sql = self.execute

    def execute(self, **kwargs):
        if type(self.value) is str:
            return f"'{self.value}' as {self.alias}" if self.alias else f"'{self.value}'" 
        elif type(self.value) in (list, tuple):
            return '(' + ', '.join([f"{item}" if type(item) in (int, float) else f"'{item}'" for item in self.value]) + ')'
        elif hasattr(self.value, 'get_sql'):
            return self.value.get_sql(**kwargs)
        else:
            return f"{self.value} as {self.alias}" if self.alias else f"{self.value}"


class NullValue(Term):
    """NULL value for use in queries."""
    def __init__(self) -> None:
        super().__init__()
        self.get_sql = self.execute

    def execute(self, **kwargs):
        return f"NULL as {self.alias}" if self.alias else f"NULL"

# %% ../nbs/02_terms.ipynb 7
class FieldBase(Term):
    """Collection of methods to convert to ArithmeticExpression and Criteria"""
    def __add__(self, other):
        return ArithmeticExpression(self, '+', other)

    def __radd__(self, other):
        return ArithmeticExpression(other, '+', self)

    def __sub__(self, other):
        return ArithmeticExpression(self, '-', other)

    def __rsub__(self, other):
        return ArithmeticExpression(other, '-', self)

    def __mul__(self, other):
        return ArithmeticExpression(self, '*', other)
        
    def __rmul__(self, other):
        return ArithmeticExpression(other, '*', self)

    def __truediv__(self, other):
        return ArithmeticExpression(self, '/', other)

    def __rtruediv__(self, other):
        return ArithmeticExpression(other, '/', self)

    def __gt__(self, other):
        return Criteria(self, '>', other)

    def __ge__(self, other):
        return Criteria(self, '>=', other)
        
    def __eq__(self, other):
        return Criteria(self, '=', other)

    def __ne__(self, other):
        return Criteria(self, '<>', other)

    def __le__(self, other):
        return Criteria(self, '<=', other)

    def __lt__(self, other):
        return Criteria(self, '<', other)

    gt = __gt__
    ge = __ge__
    eq = __eq__
    ne = __ne__
    le = __le__
    lt = __lt__

    def isnull(self):
        return Criteria(self, 'IS', NullValue())

    def notnull(self):
        return Criteria(self, 'IS NOT', NullValue())

    def like(self, other):
        return Criteria(self, 'LIKE', Value(other))
    
    def ilike(self, other):
        return Criteria(self, 'ILIKE', Value(other))

    def not_like(self, other):
        return Criteria(self, 'NOT LIKE', Value(other))

    def not_ilike(self, other):
        return Criteria(self, 'NOT ILIKE', Value(other))
    
    def isin(self, other:list):
        return Criteria(self, 'IN', Value(other))
    
    def notin(self, other):
        return Criteria(self, 'NOT IN', Value(other))

    #TODO: implement new Criteria, and execute for list


class Field(FieldBase):
    """A simple wrapper for a field that will be used in a query. Quotes can be added to the field name by setting the `quote_char` parameter in get_sql() method."""
    def __init__(self, name) -> None: 
        self.name = name
        self.alias = None
        self.get_sql = self.execute

    def quoted_name(self, quote_char):
        name_list = self.name.split('.')
        name = '.'.join([f"{quote_char}{n}{quote_char}" for n in name_list])
        return name

    def execute(self, **kwargs):
        q = kwargs.get('quote_char', '') or ''
        quoted_name = self.quoted_name(q)
        if self.alias:
            return f"{quoted_name} AS {q}{self.alias}{q}"
        else:
            return quoted_name
    

class ArithmeticExpression(FieldBase):
    """Constructor for arithmetic expressions from two terms and automatically adds parentheses in appropriate places."""
    add_order = ["+", "-"]

    def __init__(self, this, op, other) -> None:
        super().__init__()
        self.this, self.op, self.other = this, op, other
        self.alias = None
        self.get_sql = self.execute

    def left_needs_parens(self, left_op, curr_op) -> bool:
        """
        Returns true if the expression on the left of the current operator needs to be enclosed in parentheses.
        :param current_op:
            The current operator.
        :param left_op:
            The highest level operator of the left expression.
        """
        if left_op is None or curr_op in self.add_order:
            # If the left expression is a single item.
            # or if the current operator is '+' or '-'.
            return False
        
        # The current operator is '*' or '/'. 
        # If the left operator is '+' or '-', we need to add parentheses:
        # e.g. (A + B) / ..., (A - B) / ...
        # Otherwise, no parentheses are necessary:
        # e.g. A * B / ..., A / B / ...
        return left_op in self.add_order

    def right_needs_parens(self, curr_op, right_op) -> bool:
        """
        Returns true if the expression on the right of the current operator needs to be enclosed in parentheses.
        :param current_op:
            The current operator.
        :param right_op:
            The highest level operator of the right expression.
        """
        if right_op is None:
            # If the right expression is a single item.
            return False
        # If the right operator is '+' or '-', we always add parentheses:
        # e.g. ... - (A + B), ... - (A - B), ... + (A + B)
        # Otherwise, no parentheses are necessary:
        # e.g. ... - A / B, ... - A * B
        return right_op in self.add_order

    def execute(self, **kwargs):
        q = kwargs.get('quote_char', '') or ''
        if self.this.__class__ is ArithmeticExpression:
            this = self.this.execute(**kwargs)
            this = f"({this})" if self.left_needs_parens(self.this.op, self.op) else this
        elif self.this.__class__ is Field:
            this = self.this.alias or execute(self.this, **kwargs)
        else:
            this = execute(self.this, **kwargs)

        if self.other.__class__ is ArithmeticExpression:
            other = self.other.execute(**kwargs)
            other = f"({other})" if self.right_needs_parens(self.op, self.other.op) else other
        elif self.other.__class__ is Field:
            other = self.other.alias or execute(self.other, **kwargs)
        else:
            other = execute(self.other, **kwargs)

        if self.alias:
            return f"{this} {self.op} {other} AS {q}{self.alias}{q}"
        else:
            return f"{this} {self.op} {other}"


class Criteria:
    """Constructor for criteria from two terms and automatically adds parentheses in appropriate places."""
    compose_ops = ('and', 'or')

    def __init__(self, this, op, other) -> None:
        super().__init__()
        self.this, self.op, self.other = this, op, other
        self.add_parentheses = False
        self.not_ = False
        self.get_sql = self.execute

    def compose_criteria(self, op, other):
        """Add parentheses when operator in criteria is different from op"""
        if other.__class__ is Criteria:
            if self.op in self.compose_ops and op in self.compose_ops and self.op != op:
                self.add_parentheses = True
            if other.op in self.compose_ops and op in self.compose_ops and other.op != op:
                other.add_parentheses = True
        return Criteria(self, op, other)

    @staticmethod
    def resolve(obj, **kwargs):
        if obj.__class__ is Criteria:
            obj_c = obj.execute(**kwargs)
            obj_c = f"({obj_c})" if obj.add_parentheses == True else obj_c
        else:
            obj_c = execute(obj, **kwargs)
        return obj_c
    
    def execute(self, **kwargs):
        this = self.resolve(self.this, **kwargs)
        other = self.resolve(self.other, **kwargs)
        self.result =  f"{this} {self.op} {other}"
        if self.not_:
            return f"NOT ({self.result})"
        else:
            return self.result
    
    def negate(self):
        self.not_ = True
        return self

    def __and__(self, __o):
        return self.compose_criteria('and', __o)

    def __or__(self, __o):
        return self.compose_criteria('or', __o)

# %% ../nbs/02_terms.ipynb 10
def kwargs_func(func, *args, **kwargs):
    "Allow arbitrary kwargs. Only pass those kwargs that are specified in func to func."
    sig = inspect.signature(func)
    param = sig.parameters
    func_kwargs = {k:v for k, v in param.items() if v.default!=inspect._empty}
    kwargs = {k:v for k, v in kwargs.items() if k in func_kwargs}
    return func(*args, **kwargs)


class DelayedFunc(FieldBase):
    """Delay the execution of stored function until exec is run."""
    def __init__(self, 
                 func, 
                 args,
                 kwargs, 
                 window_func=True # whether this function is a window function
                 ) -> None: 
        super().__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.window_func = window_func
        self.get_sql = self.execute
    

    def execute(self, **kwargs):
        """keyword arguments can be overwritten with any provided new kwargs."""
        self.kwargs.update(kwargs)
        # recursively resolve all delayed functions
        args = (execute(arg, **self.kwargs) for arg in self.args) 
        func = self.func(*args, **self.kwargs)
        if self.alias:
            return f"{func} AS {self.alias}"
        else:
            return func


def custom_func(func=None, window_func=False, dialect=None):
    """return Field"""
    if func is None: 
        return partial(custom_func, window_func=window_func, dialect=dialect)
    else:
        if dialect is None:
            @delegates(func)
            def wrapper(*args, **kwargs):
                dlf = DelayedFunc(func, args, kwargs, window_func=window_func)
                return dlf
        else:
            # get previously defined func
            func_name = func.__name__
            ori_func = globals().get(func_name)
            @delegates(func)
            def wrapper(*args, **kwargs):
                def new_func(*args, **kwargs):
                    # if the dialect is different from the one defined for this new function, fall back to the original func
                    if kwargs['dialect'] != dialect and ori_func:
                        f = ori_func().func
                    else:
                        f = func
                    return kwargs_func(f, *args, **kwargs)

                # make new delayed function
                dlf = DelayedFunc(new_func, args, kwargs, window_func=window_func)
                return dlf
        return wrapper

# %% ../nbs/02_terms.ipynb 16
class OverClause:
    """Constructor for OVER clause."""
    def __init__(self, expression) -> None:
        if not (
            (hasattr(expression, 'window_func')
             and 
             expression.window_func==True)
            or 
            (type(expression) is str)
        ):
            raise ValueError(f"Expression has to be of the ArithmeticExpression type or a Window Function (DelayedFunc with window_func=True)!!")
        self.expr = expression
        self.alias = None
        self.rows_flag = False
        self.range_flag = False
        self.d = {}
        self.get_sql = self.execute

    def over(self, q):
        self.d['PARTITION BY'] = q
        return self

    def orderby(self, q):
        self.d['ORDER BY'] = q
        return self

    def _check_rows_or_range(self):
        if self.rows_flag==True: 
            raise ValueError(f"ROWS already set!")
        if self.range_flag==True:
            raise ValueError(f"RANGE already set!")

    def rows(self, start, end):
        self._check_rows_or_range()
        self.rows = True
        self.d['ROWS'] = (start, end)
        return self

    def range(self, start, end):
        self._check_rows_or_range()
        self.range = True
        self.d['RANGE'] = (start, end)
        return self

    def as_(self, alias):
        self.alias = alias
        return self

    def _resolve_over_statement(self, **kwargs):
        sql = []
        for k in ['PARTITION BY', 'ORDER BY', 'ROWS', 'RANGE']:
            if k in self.d:
                if k in ['PARTITION BY', 'ORDER BY']:
                    rslvd = f"{k} {execute(self.d[k], **kwargs)}"
                else:
                    start, end = self.d[k]
                    rslvd = f"{k} BETWEEN {execute(start, **kwargs)} AND {execute(end, **kwargs)}"
                sql.append(rslvd)
        return f"{execute(self.expr, **kwargs)} OVER ({' '.join(sql)})"

    def execute(self, **kwargs):
        sql = self._resolve_over_statement(**kwargs)
        if self.alias:
            return f"{sql} AS {self.alias}"
        else:
            return sql


class Preceding:
    def __init__(self, N=None) -> None:
        self.N = N
        self.get_sql = self.execute

    def execute(self, **kwargs):
        if self.N:
            return f"{self.N} PRECEDING"
        else:
            return "UNBOUNDED PRECEDING"


class Following:
    def __init__(self, N=None) -> None:
        self.N = N
        self.get_sql = self.execute

    def execute(self, **kwargs):
        if self.N:
            return f"{self.N} FOLLOWING"
        else:
            return "UNBOUNDED FOLLOWING"


CURRENT_ROW = "CURRENT_ROW"

# %% ../nbs/02_terms.ipynb 17
@patch
def over(self:ArithmeticExpression, partition_by):
    return OverClause(self).over(partition_by)

@patch
def over(self:DelayedFunc, partition_by):
    if self.window_func==True:
        return OverClause(self).over(partition_by)
    else:
        raise ValueError(f"This function is not a window function!!")


# %% ../nbs/02_terms.ipynb 20
class Case(Term):
    """Constructor for CASE statement."""
    def __init__(self) -> None:
        self.dp = []
        self.alias = None
        self.get_sql = self.execute

    def check_prev(self, statement):
        if self.dp:
            prev = self.dp[0][0]
            if prev == statement:
                return True
        return False

    def when(self, q, then):
        if self.check_prev('ELSE'):
            raise ValueError(f"'WHEN' can not follow 'ELSE'!")
        self.dp.append(('WHEN', q, then))
        return self

    def else_(self, q):
        self.dp.append(('ELSE', q))
        return self

    def execute(self, **kwargs):
        sql = ["CASE"]
        for item in self.dp:
            if item[0] == 'WHEN':
                q_resolved = f"WHEN {execute(item[1], **kwargs)} THEN {execute(item[2], **kwargs)}"
            else:
                q_resolved = f"ELSE {execute(item[1], **kwargs)}"
            sql.append(q_resolved)
        if self.alias:
            sql.append(f"END AS {self.alias}")
        else:
            sql.append("END")
        return '\n'.join(sql)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_utils.ipynb.

# %% auto 0
__all__ = ['quote_symbol', 'to_sql', 'execute', 'delegates']

# %% ../nbs/01_utils.ipynb 2
def quote_symbol(quote):
    """generate quote symbol to use for tables and columns"""
    if type(quote) == str:
        return quote
    else:
        if quote == True:
            return '"'
        else:
            return ''


def to_sql(l):
    if type(l) == list:
        return '(' + ', '.join([f"{item}" if type(item) in (int, float) else f"'{item}'" for item in l]) + ')'
    else:
        raise ValueError(f"type {type(l)} for l is not implemented!")
    return l


def execute(obj, **kwargs):
    """Call the exec method of `obj` if it exists, otherwise return the string representation of `obj`"""
    if hasattr(obj, 'execute'):
        return obj.execute(**kwargs)
    else:
        return obj

# %% ../nbs/01_utils.ipynb 4
def delegates(to):
    """Decorator: replace func name, signature and docstring with `to`'s."""
    import inspect
    from fastcore.docments import qual_name
    def _f(f):
        f.__name__ = to.__name__
        f.__qualname__ = qual_name(to)
        f.__signature__ = inspect.signature(to)
        f.__doc__ = to.__doc__
        return f
    return _f
